@@ -141,39 +141,6 @@ int disk_read_sectors(struct volume *volume, void *buf, uint64_t block, size_t c
     return DISK_SUCCESS;
 }
 
-static int disk_write_sectors(struct volume *volume, void *buf, uint64_t block, size_t count) {
-    struct dap dap = {0};
-
-    if (count * volume->sector_size > XFER_BUF_SIZE)
-        panic(false, "XFER");
-
-    if (xfer_buf == NULL)
-        xfer_buf = conv_mem_alloc(XFER_BUF_SIZE);
-
-    dap.size    = 16;
-    dap.count   = count;
-    dap.segment = rm_seg(xfer_buf);
-    dap.offset  = rm_off(xfer_buf);
-    dap.lba     = block;
-
-    struct rm_regs r = {0};
-    r.eax = 0x4301;
-    r.edx = volume->drive;
-    r.esi = (uint32_t)rm_off(&dap);
-    r.ds  = rm_seg(&dap);
-
-    if (buf != NULL)
-        memcpy(xfer_buf, buf, count * volume->sector_size);
-
-    rm_int(0x13, &r, &r);
-
-    if (r.eflags & EFLAGS_CF) {
-        return DISK_FAILURE;
-    }
-
-    return DISK_SUCCESS;
-}
-
 static bool detect_sector_size(struct volume *volume) {
     struct dap dap = {0};
 
@@ -311,11 +278,9 @@ void disk_create_index(void) {
             continue;
         }
 
-        if (disk_read_sectors(block, xfer_buf, 0, 1) != DISK_SUCCESS) {
-            continue;
-        }
-
-        block->is_optical = (disk_write_sectors(block, xfer_buf, 0, 1) != DISK_SUCCESS) && block->sector_size == 2048;
+        // Detect optical drives via DPTE ATAPI bit (bit 6) or sector size heuristic
+        bool is_atapi = (dpte != NULL && (dpte->flags & (1 << 6)));
+        block->is_optical = is_atapi || (block->sector_size == 2048 && is_removable);
 
         if (!is_removable && !block->is_optical) {
             if (consumed_bda_disks == bda_disk_count) {
@@ -484,7 +449,7 @@ static bool is_efi_handle_to_skip(EFI_HANDLE efi_handle) {
     return false;
 }
 
-static bool is_efi_handle_hdd(EFI_HANDLE efi_handle) {
+static bool is_efi_handle_optical(EFI_HANDLE efi_handle) {
     EFI_STATUS status;
 
     EFI_GUID dp_guid = EFI_DEVICE_PATH_PROTOCOL_GUID;
@@ -500,11 +465,8 @@ static bool is_efi_handle_hdd(EFI_HANDLE efi_handle) {
             break;
         }
 
-        if (dp->Type == MEDIA_DEVICE_PATH) {
-            switch (dp->SubType) {
-                case MEDIA_HARDDRIVE_DP:
-                    return true;
-            }
+        if (dp->Type == MEDIA_DEVICE_PATH && dp->SubType == MEDIA_CDROM_DP) {
+            return true;
         }
 
         uint16_t len = *(uint16_t *)dp->Length;
@@ -517,9 +479,119 @@ static bool is_efi_handle_hdd(EFI_HANDLE efi_handle) {
     return false;
 }
 
-struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
-    struct volume *ret;
+static EFI_DEVICE_PATH_PROTOCOL *get_device_path(EFI_HANDLE efi_handle) {
+    EFI_STATUS status;
+    EFI_GUID dp_guid = EFI_DEVICE_PATH_PROTOCOL_GUID;
+    EFI_DEVICE_PATH_PROTOCOL *dp = NULL;
 
+    status = gBS->HandleProtocol(efi_handle, &dp_guid, (void **)&dp);
+    if (status) {
+        return NULL;
+    }
+    return dp;
+}
+
+// Compare device paths up to (but not including) partition nodes
+static bool device_paths_match_disk(EFI_DEVICE_PATH_PROTOCOL *dp1,
+                                    EFI_DEVICE_PATH_PROTOCOL *dp2) {
+    if (dp1 == NULL || dp2 == NULL) {
+        return false;
+    }
+
+    while (!IsDevicePathEnd(dp1) && !IsDevicePathEnd(dp2)) {
+        // Stop at partition nodes
+        if (dp1->Type == MEDIA_DEVICE_PATH &&
+            (dp1->SubType == MEDIA_HARDDRIVE_DP || dp1->SubType == MEDIA_CDROM_DP)) {
+            break;
+        }
+        if (dp2->Type == MEDIA_DEVICE_PATH &&
+            (dp2->SubType == MEDIA_HARDDRIVE_DP || dp2->SubType == MEDIA_CDROM_DP)) {
+            break;
+        }
+
+        uint16_t len1 = DevicePathNodeLength(dp1);
+        uint16_t len2 = DevicePathNodeLength(dp2);
+
+        if (len1 != len2) {
+            return false;
+        }
+
+        if (len1 < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+            return false;
+        }
+
+        if (memcmp(dp1, dp2, len1) != 0) {
+            return false;
+        }
+
+        dp1 = (void *)dp1 + len1;
+        dp2 = (void *)dp2 + len2;
+    }
+
+    return true;
+}
+
+static struct volume *volume_by_device_path(EFI_HANDLE query_handle) {
+    EFI_DEVICE_PATH_PROTOCOL *query_dp = get_device_path(query_handle);
+    if (query_dp == NULL) {
+        return NULL;
+    }
+
+    for (size_t i = 0; i < volume_index_i; i++) {
+        EFI_DEVICE_PATH_PROTOCOL *vol_dp = get_device_path(volume_index[i]->efi_handle);
+        if (vol_dp == NULL) {
+            continue;
+        }
+
+        if (device_paths_match_disk(query_dp, vol_dp)) {
+            // Convert first_sect from 512-byte sectors to device LBAs
+            int sector_size = volume_index[i]->sector_size;
+            if ((volume_index[i]->first_sect * 512) % sector_size) {
+                continue;  // Misaligned, skip this volume
+            }
+            uint64_t first_sect_lba = (volume_index[i]->first_sect * 512) / sector_size;
+
+            EFI_DEVICE_PATH_PROTOCOL *qp = query_dp;
+            while (!IsDevicePathEnd(qp)) {
+                if (qp->Type == MEDIA_DEVICE_PATH && qp->SubType == MEDIA_HARDDRIVE_DP) {
+                    uint16_t len = DevicePathNodeLength(qp);
+                    if (len < sizeof(HARDDRIVE_DEVICE_PATH)) {
+                        break;
+                    }
+                    HARDDRIVE_DEVICE_PATH *query_hd = (HARDDRIVE_DEVICE_PATH *)qp;
+                    if (first_sect_lba == query_hd->PartitionStart) {
+                        return volume_index[i];
+                    }
+                    break;
+                }
+                if (qp->Type == MEDIA_DEVICE_PATH && qp->SubType == MEDIA_CDROM_DP) {
+                    uint16_t len = DevicePathNodeLength(qp);
+                    if (len < sizeof(CDROM_DEVICE_PATH)) {
+                        break;
+                    }
+                    CDROM_DEVICE_PATH *query_cd = (CDROM_DEVICE_PATH *)qp;
+                    if (first_sect_lba == query_cd->PartitionStart) {
+                        return volume_index[i];
+                    }
+                    break;
+                }
+                uint16_t len = DevicePathNodeLength(qp);
+                if (len < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+                    break;
+                }
+                qp = (void *)qp + len;
+            }
+
+            if (IsDevicePathEnd(qp) && volume_index[i]->partition == 0) {
+                return volume_index[i];
+            }
+        }
+    }
+
+    return NULL;
+}
+
+struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
     EFI_STATUS status;
 
     EFI_GUID block_io_guid = BLOCK_IO_PROTOCOL;
@@ -535,118 +607,23 @@ struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
     }
 
     uint64_t bdev_size = ((uint64_t)block_io->Media->LastBlock + 1) * (uint64_t)block_io->Media->BlockSize;
-    if (bdev_size < UNIQUE_SECTOR_POOL_SIZE) {
-        goto fallback;
-    }
+    if (bdev_size >= UNIQUE_SECTOR_POOL_SIZE) {
+        status = block_io->ReadBlocks(block_io, block_io->Media->MediaId,
+                                      0,
+                                      UNIQUE_SECTOR_POOL_SIZE,
+                                      unique_sector_pool);
+        if (status == 0) {
+            uint8_t b2b[BLAKE2B_OUT_BYTES];
+            blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
 
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId,
-                                  0,
-                                  UNIQUE_SECTOR_POOL_SIZE,
-                                  unique_sector_pool);
-    if (status != 0) {
-        goto fallback;
-    }
-
-    uint8_t b2b[BLAKE2B_OUT_BYTES];
-    blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
-
-    ret = volume_by_unique_sector(b2b);
-    if (ret != NULL) {
-        return ret;
-    }
-
-    // Fallback to read-back method
-fallback:;
-    if (!is_efi_handle_hdd(efi_handle)) {
-        return NULL;
-    }
-
-    uint64_t signature = rand64();
-    uint64_t new_signature;
-    do { new_signature = rand64(); } while (new_signature == signature);
-    uint64_t orig;
-
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    orig = *(uint64_t *)unique_sector_pool;
-    if (status) {
-        return NULL;
-    }
-
-    *(uint64_t *)unique_sector_pool = signature;
-    status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        // Attempt to restore anyways, just in case.
-        *(uint64_t *)unique_sector_pool = orig;
-        block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-        return NULL;
-    }
-
-    ret = NULL;
-    for (size_t i = 0; i < volume_index_i; i++) {
-        uint64_t compare;
-
-        status = volume_index[i]->block_io->ReadBlocks(volume_index[i]->block_io,
-                          volume_index[i]->block_io->Media->MediaId,
-                          (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size,
-                          4096, unique_sector_pool);
-        compare = *(uint64_t *)unique_sector_pool;
-        if (status) {
-            continue;
-        }
-
-        if (compare == signature) {
-            // Double check
-            status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-            *(uint64_t *)unique_sector_pool = new_signature;
-            status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-
-            status = volume_index[i]->block_io->ReadBlocks(volume_index[i]->block_io,
-                          volume_index[i]->block_io->Media->MediaId,
-                          (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size,
-                          4096, unique_sector_pool);
-            compare = *(uint64_t *)unique_sector_pool;
-            if (status) {
-                continue;
-            }
-
-            if (compare == new_signature) {
-                ret = volume_index[i];
-                break;
-            }
-
-            status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-            *(uint64_t *)unique_sector_pool = signature;
-            status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
+            struct volume *ret = volume_by_unique_sector(b2b);
+            if (ret != NULL) {
+                return ret;
             }
         }
     }
 
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        return NULL;
-    }
-    *(uint64_t *)unique_sector_pool = orig;
-    status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        return NULL;
-    }
-
-    if (ret != NULL) {
-        return ret;
-    }
-
-    return NULL;
+    return volume_by_device_path(efi_handle);
 }
 
 static void find_unique_sectors(void) {
@@ -751,20 +728,12 @@ fail:
         if (drive->Media->LogicalPartition)
             continue;
 
-        drive->Media->WriteCaching = false;
-
-        status = drive->ReadBlocks(drive, drive->Media->MediaId, 0, 4096, unique_sector_pool);
-        if (status) {
-            continue;
-        }
-
-        status = drive->WriteBlocks(drive, drive->Media->MediaId, 0, 4096, unique_sector_pool);
-
-        drive->Media->WriteCaching = true;
-
         struct volume *block = ext_mem_alloc(sizeof(struct volume));
 
-        if ((status || drive->Media->ReadOnly) && drive->Media->BlockSize == 2048) {
+        bool is_optical = is_efi_handle_optical(handles[i]) ||
+                          (drive->Media->ReadOnly && drive->Media->BlockSize == 2048);
+
+        if (is_optical) {
             block->index = optical_indices++;
             block->is_optical = true;
         } else {
