@@ -115,11 +115,10 @@ static void iso9660_find_PVD(struct iso9660_primary_volume *desc, struct volume
     uint32_t max_lba = ISO9660_FIRST_VOLUME_DESCRIPTOR + ISO9660_MAX_VOLUME_DESCRIPTORS;
 
     while (lba < max_lba) {
-        uint64_t offset;
-        if (__builtin_mul_overflow((uint64_t)lba, (uint64_t)ISO9660_SECTOR_SIZE, &offset)) {
-            panic(false, "ISO9660: volume descriptor offset overflow");
+        uint64_t offset = (uint64_t)lba * ISO9660_SECTOR_SIZE;
+        if (!volume_read(vol, desc, offset, sizeof(struct iso9660_primary_volume))) {
+            panic(false, "ISO9660: failed to read volume descriptor");
         }
-        volume_read(vol, desc, offset, sizeof(struct iso9660_primary_volume));
 
         switch (desc->volume_descriptor.type) {
         case ISO9660_VDT_PRIMARY:
@@ -149,11 +148,10 @@ static void iso9660_cache_root(struct volume *vol,
     }
 
     *root = ext_mem_alloc(*root_size);
-    uint64_t offset;
-    if (__builtin_mul_overflow((uint64_t)pv.root.extent.little, (uint64_t)ISO9660_SECTOR_SIZE, &offset)) {
-        panic(false, "ISO9660: root extent offset overflow");
+    uint64_t offset = (uint64_t)pv.root.extent.little * ISO9660_SECTOR_SIZE;
+    if (!volume_read(vol, *root, offset, *root_size)) {
+        panic(false, "ISO9660: failed to read root directory");
     }
-    volume_read(vol, *root, offset, *root_size);
 }
 
 static struct iso9660_context *iso9660_get_context(struct volume *vol) {
@@ -193,7 +191,7 @@ static bool load_name(char *buf, size_t limit, struct iso9660_directory_entry *e
 
     int rrnamelen = 0;
     unsigned char *nm_entry = NULL;
-    while ((sysarea_len >= 4) && ((sysarea[3] == 1) || (sysarea[2] == 2))) {
+    while ((sysarea_len >= 4) && (sysarea[3] == 1)) {
         // Validate entry length doesn't exceed remaining sysarea
         if (sysarea[2] > sysarea_len) {
             break;
@@ -352,7 +350,9 @@ static void iso9660_close(struct file_handle *file);
 struct file_handle *iso9660_open(struct volume *vol, const char *path) {
     char buf[6];
     const uint64_t signature = ISO9660_FIRST_VOLUME_DESCRIPTOR * ISO9660_SECTOR_SIZE + 1;
-    volume_read(vol, buf, signature, 5);
+    if (!volume_read(vol, buf, signature, 5)) {
+        return NULL;
+    }
     buf[5] = '\0';
     if (strcmp(buf, "CD001") != 0) {
         return NULL;
@@ -482,13 +482,12 @@ struct file_handle *iso9660_open(struct volume *vol, const char *path) {
 
         first = false;
 
-        uint64_t dir_offset;
-        if (__builtin_mul_overflow((uint64_t)next_sector, (uint64_t)ISO9660_SECTOR_SIZE, &dir_offset)) {
+        uint64_t dir_offset = (uint64_t)next_sector * ISO9660_SECTOR_SIZE;
+        if (!volume_read(vol, current, dir_offset, current_size)) {
             pmm_free(current, current_size);
             pmm_free(ret, sizeof(struct iso9660_file_handle));
             return NULL;
         }
-        volume_read(vol, current, dir_offset, current_size);
     }
 
     // Free the last directory buffer if we allocated one (not the cached root)
@@ -533,16 +532,11 @@ static void iso9660_read(struct file_handle *file, void *buf, uint64_t loc, uint
             uint64_t bytes_available = extent_size - offset_in_extent;
             uint64_t to_read = (count < bytes_available) ? count : bytes_available;
 
-            uint64_t base_offset;
-            if (__builtin_mul_overflow((uint64_t)f->extents[i].LBA, (uint64_t)ISO9660_SECTOR_SIZE, &base_offset)) {
-                panic(false, "iso9660: offset calculation overflow");
-            }
-            uint64_t disk_offset;
-            if (__builtin_add_overflow(base_offset, offset_in_extent, &disk_offset)) {
-                panic(false, "iso9660: offset calculation overflow");
-            }
+            uint64_t disk_offset = (uint64_t)f->extents[i].LBA * ISO9660_SECTOR_SIZE + offset_in_extent;
 
-            volume_read(f->context->vol, buf, disk_offset, to_read);
+            if (!volume_read(f->context->vol, buf, disk_offset, to_read)) {
+                panic(false, "iso9660: failed to read file data");
+            }
 
             buf = (uint8_t *)buf + to_read;
             loc += to_read;
