@@ -95,10 +95,8 @@ struct smbios_entry_point_64 {
     uint8_t docrev;
     uint8_t entry_point_revision;
     uint8_t reserved;
-    /// Size of the largest SMBIOS structure, in bytes, and encompasses the
-    /// structure’s formatted area and text strings
-    uint16_t max_structure_size;
-    uint16_t padding;
+    /// Size of the SMBIOS Structure Table, in bytes.
+    uint32_t table_maximum_size;
     /// 64-bit physical starting address of the read-only SMBIOS Structure
     /// Table.
     uint64_t table_address;
