@@ -385,10 +385,14 @@ __attribute__((always_inline)) static inline void genloop(struct fb_info *fb, si
             }
             else { /* internal part */
                 for (size_t x = xstart; x < xend; x++) {
-                    size_t image_x = (x - background->x_displacement);
-                    bool x_external = image_x >= background->x_size;
-                    uint32_t img_pixel = *(uint32_t*)(img + image_x * colsize + off);
-                    uint32_t i = blend(fb, x, y, x_external ? background->back_colour : img_pixel);
+                    uint32_t pixel;
+                    if (x < background->x_displacement || x - background->x_displacement >= background->x_size) {
+                        pixel = background->back_colour;
+                    } else {
+                        size_t image_x = x - background->x_displacement;
+                        pixel = *(uint32_t*)(img + image_x * colsize + off);
+                    }
+                    uint32_t i = blend(fb, x, y, pixel);
                     bg_canvas[canvas_off + x] = i;
                 }
             }
@@ -426,6 +430,10 @@ static void loop_internal(struct fb_info *fb, size_t xstart, size_t xend, size_t
 
 static void generate_canvas(struct fb_info *fb) {
     if (background) {
+        // Free previous canvas if it exists
+        if (bg_canvas != NULL) {
+            pmm_free(bg_canvas, bg_canvas_size);
+        }
         bg_canvas_size = fb->framebuffer_width * fb->framebuffer_height * sizeof(uint32_t);
         bg_canvas = ext_mem_alloc(bg_canvas_size);
 
@@ -665,6 +673,8 @@ bool gterm_init(struct fb_info **_fbs, size_t *_fbs_count,
     char *menu_font_size = config_get_value(config, 0, "TERM_FONT_SIZE");
     if (menu_font_size != NULL) {
         parse_resolution(&tmp_font_width, &tmp_font_height, NULL, menu_font_size);
+        // XXX: Do not allow widths != 8
+        tmp_font_width = 8;
 
         size_t tmp_font_size = (tmp_font_width * tmp_font_height * FLANTERM_FB_FONT_GLYPHS) / 8;
 
@@ -682,6 +692,11 @@ bool gterm_init(struct fb_info **_fbs, size_t *_fbs_count,
         if ((f = uri_open(menu_font)) == NULL) {
             print("menu: Could not open font file.\n");
         } else {
+            if (font_size > f->size) {
+                print("Font size too large for provided font file. Not loading.\n");
+                fclose(f);
+                goto no_load_font;
+            }
             fread(f, font, 0, font_size);
             if (menu_font_size != NULL) {
                 font_width = tmp_font_width;
@@ -763,8 +778,8 @@ no_load_font:;
             fb->framebuffer_height = tmp;
         }
 
-        terms[terms_i] = flanterm_fb_init(ext_mem_alloc,
-                            pmm_free,
+        terms[terms_i] = flanterm_fb_init(ext_mem_alloc_size_t,
+                            pmm_free_size_t,
                             (void *)(uintptr_t)fb->framebuffer_addr,
                             fb->framebuffer_width, fb->framebuffer_height, fb->framebuffer_pitch,
                             fb->red_mask_size, fb->red_mask_shift,
