@@ -73,8 +73,18 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
     struct multiboot1_header header = {0};
     size_t header_offset = 0;
 
-    for (header_offset = 0; header_offset < 8192; header_offset += 4) {
-        uint32_t v = *(uint32_t *)(kernel+header_offset);
+    // Per Multiboot spec, header must be within first 8192 bytes and 4-byte aligned.
+    // Ensure we don't read past end of file when checking magic or copying header.
+    size_t search_limit = 8192;
+    if (kernel_file_size < sizeof(struct multiboot1_header)) {
+        panic(true, "multiboot1: Kernel file too small to contain header");
+    }
+    if (search_limit > kernel_file_size - sizeof(struct multiboot1_header)) {
+        search_limit = kernel_file_size - sizeof(struct multiboot1_header);
+    }
+
+    for (header_offset = 0; header_offset <= search_limit; header_offset += 4) {
+        uint32_t v = *(uint32_t *)(kernel + header_offset);
 
         if (v == MULTIBOOT1_HEADER_MAGIC) {
             memcpy(&header, kernel + header_offset, sizeof(header));
@@ -100,11 +110,20 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
         if (header.load_addr > header.header_addr)
             panic(true, "multiboot1: Illegal load address");
 
+        size_t addr_diff = header.header_addr - header.load_addr;
+        if (addr_diff > header_offset)
+            panic(true, "multiboot1: Address tag offset underflow");
+
+        size_t load_src = header_offset - addr_diff;
+
         size_t load_size;
-        if (header.load_end_addr)
+        if (header.load_end_addr) {
             load_size = header.load_end_addr - header.load_addr;
-        else
-            load_size = kernel_file_size;
+        } else {
+            if (load_src > kernel_file_size)
+                panic(true, "multiboot1: Load source exceeds kernel file size");
+            load_size = kernel_file_size - load_src;
+        }
 
         uint32_t bss_size = 0;
         if (header.bss_end_addr) {
@@ -119,8 +138,7 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
 
         void *elsewhere = ext_mem_alloc(full_size);
 
-        memcpy(elsewhere, kernel + (header_offset
-                - (header.header_addr - header.load_addr)), load_size);
+        memcpy(elsewhere, kernel + load_src, load_size);
 
         entry_point = header.entry_addr;
 
@@ -206,16 +224,21 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
         mb1_info_alloc(&mb1_info_raw, sizeof(struct multiboot1_info));
 
     if (section_hdr_info_valid == true) {
+        size_t section_table_size = (size_t)section_hdr_info.section_entry_size * section_hdr_info.num;
+        if (section_hdr_info.section_offset > kernel_file_size ||
+            section_table_size > kernel_file_size - section_hdr_info.section_offset) {
+            panic(true, "multiboot1: ELF section headers out of bounds");
+        }
+
         multiboot1_info->elf_sect.num = section_hdr_info.num;
         multiboot1_info->elf_sect.size = section_hdr_info.section_entry_size;
         multiboot1_info->elf_sect.shndx = section_hdr_info.str_section_idx;
 
-        void *sections = mb1_info_alloc(&mb1_info_raw,
-            section_hdr_info.section_entry_size * section_hdr_info.num);
+        void *sections = mb1_info_alloc(&mb1_info_raw, section_table_size);
 
         multiboot1_info->elf_sect.addr = (uintptr_t)sections - mb1_info_slide;
 
-        memcpy(sections, kernel + section_hdr_info.section_offset, section_hdr_info.section_entry_size * section_hdr_info.num);
+        memcpy(sections, kernel + section_hdr_info.section_offset, section_table_size);
 
         int bits = elf_bits(kernel);
 
@@ -227,6 +250,11 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
                     continue;
                 }
 
+                if (shdr->sh_offset > kernel_file_size ||
+                    shdr->sh_size > kernel_file_size - shdr->sh_offset) {
+                    continue;
+                }
+
                 uint64_t section = (uint64_t)-1; /* no target preference, use top */
 
                 if (!elsewhere_append(true /* flexible target */,
@@ -243,6 +271,11 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
                     continue;
                 }
 
+                if (shdr->sh_offset > kernel_file_size ||
+                    shdr->sh_size > kernel_file_size - shdr->sh_offset) {
+                    continue;
+                }
+
                 uint64_t section = (uint64_t)-1; /* no target preference, use top */
 
                 if (!elsewhere_append(true /* flexible target */,
@@ -250,7 +283,7 @@ noreturn void multiboot1_load(char *config, char *cmdline) {
                         kernel + shdr->sh_offset, &section, shdr->sh_size)) {
                     panic(true, "multiboot1: Cannot allocate elf sections");
                 }
-                        
+
                 shdr->sh_addr = section;
             }
         }
