@@ -41,7 +41,10 @@ static const char *exception_names[] = {
 
 void except(uint32_t exception, uint32_t error_code, uint32_t ebp, uint32_t eip) {
     (void)ebp;
-    panic(false, "%s exception at %x. Error code: %x", exception_names[exception], eip, error_code);
+    const char *exception_name = exception < SIZEOF_ARRAY(exception_names)
+                                 ? exception_names[exception]
+                                 : "Unknown";
+    panic(false, "%s exception at %x. Error code: %x", exception_name, eip, error_code);
 }
 
 #endif
